# Prosjektet
Dette prosjektet er et proof of concept (PoC) på hvordan Statens Vegvesen kan implementere MLOps på Google Cloud Platform, ved bruk av pipelines.
Denne pipelinen er bygget på rammeverket TFX - som er skreddersydd og eksklusiv for modeller laget med TensorFlow. 

### Forutsetninger
- Python >= 3.7, <3.9
- Google Cloud SDK >= 324.0.0
- Kubectl
- Prosjektmiljø i GCP (med tilgangsrettigheter)

### Oppsett av PoC
Under følger to metoder for å sette opp PoC. Den anbefalte fremgangsmåten er å bruke Jupyter Notebooks, da denne allerede inneholder nødvendige programmer og konfigurasjon.

## Kom i gang (Jupyter Notebook, anbefalt)
### Deploy et nettverk
Følg instruksjonene i [Opprett nettverk](#opprett-nettverk) og fortsett her etterpå.

### Opprett Notebook
Google Cloud Console: Gå til *AI Platform > Notebooks* i GCP-prosjektet du vil kjøre pipelinen i og opprett ny instans

<img src="media/notebook/notebooks1.png" width="500">

Velg Tensorflow Enterprise 2.3, uten GPU

<img src="media/notebook/notebooks2.png" width="500">

Gi den et navn, velg *europe-west3* og *europe-west3-a*, med subnettverket opprettet tidligere, *default*

<img src="media/notebook/notebooks3.png" width="500">

Åpne notebooken når den er ferdig konfigurert

<img src="media/notebook/notebooks4.png" width="500">

### Klon prosjektet
Gå til *Git > Clone a Repository*, legg inn lenke til repository og logg inn med brukernavn og passord

<img src="media/notebook/git.png" width="500">

Gå til riktig mappe

    cd bachelor2021/template/
    
### Konfigurere gcloud
Aktiver nødvendige APIer

    gcloud services enable compute.googleapis.com container.googleapis.com dataflow.googleapis.com ml.googleapis.com
    
Google Cloud Console: Gi nødvendige roller til servicekontoen "Compute Engine Default". Navigér til *IAM & Admin > Service Accounts* og trykk på service accounten

<img src="media/roles/serviceaccount.png" width="500">

Gå til "Permissions" og velg "Grant Access"

<img src="media/roles/grantaccess.png" width="500">

Velg riktig service account og rollen *Service Agents > Cloud Dataflow Service Agent* og lagre

<img src="media/roles/dataflow.png" width="500">
    
Gå videre med [Konfigurering av pipeline](#konfigurering-av-pipeline)

## Kom i gang (Lokal datamaskin, Windows er ikke anbefalt)
Oppsettet gjøres hovedsakelig gjennom kommandolinjeverktøy CMD (Windows), Terminal (Mac) eller Bash (Linux).
Følgende stil brukes på kommandoer som skal kjøres:

    Dette er en kommando som skal kjøres i kommandolinjeverktøy

### Klon prosjektet
Naviger til et område på datamaskinen du ønsker at prosjektet skal kjøre i. Last ned og installer Git dersom du ikke allerede har det og klon prosjektet. Git kan hentes fra https://git-scm.com/downloads.

### Sette opp Pythonmiljø
Sjekk om du allerede har Python (3.7 eller 3.8) installert

    python -V

eller

    python3 -V

Hvis ikke kan Python lastes ned fra https://www.python.org/downloads/. Følg
installasjonsveiledning tilhørende ditt operativsystem. **KUN Python 3.7 eller 3.8**

Navigér inn i prosjektetmappen *template*

    cd template

Opprett et virtuelt Pythonmiljø (sørg for at versjonen av Python er >= 3.7, <3.9)

    python -m venv venv

Eventuelt

    python3 -m venv venv

Aktiver det viruelle miljøet

**Mac/Linux**

    source venv/bin/activate

**Windows**

    venv\Scripts\activate

Installér nødvendige Python-pakker

    pip install -r python_requirements.txt

**NB hvis Windows:** Hvis noen pakker får feilmelding, må disse installeres manuelt uten versjonsnummer. Kjør så kommandoen under til det ikke er noen feilmeldinger.

    pip install -r python_requirements.txt

### Google Cloud SDK
Sjekk om du allerede har gcloud >= 324.0.0 installert

    gcloud --version

Hvis ikke kan det lastes ned fra https://cloud.google.com/sdk/docs/install. Velg metode for installering tilsvarende ditt operativsystem.

Sett opp autentisering mot GCP og ønsket prosjekt. En nettleser vil automatisk åpnes. Logg inn her.

    gcloud init

Sett opp autentisering mot bruk av applikasjon. En nettleser vil automatisk åpnes. Logg inn her.

    gcloud auth application-default login

Sett prosjektet du ønsker å jobbe i. Husk å bytte "myproject" med ditt prosjekt.

    gcloud config set project myproject

Aktivér GCPs API for Compute Engine, Kubernetes Engine, Dataflow og AI Platform Training & Prediction

    gcloud services enable compute.googleapis.com container.googleapis.com dataflow.googleapis.com ml.googleapis.com

Google Cloud Console: Gi nødvendige roller til servicekontoen "Compute Engine Default". Navigér til *IAM & Admin > Service Accounts* og trykk på service accounten

<img src="media/roles/serviceaccount.png" width="500">

Gå til "Permissions" og velg "Grant Access"

<img src="media/roles/grantaccess.png" width="500">

Velg riktig service account og rollen *Service Agents > Cloud Dataflow Service Agent* og lagre

<img src="media/roles/dataflow.png" width="500">

### Kubectl
Kubectl er et kommandolinjeverktøy for Kubernetes.
Installeres fra https://kubernetes.io/docs/tasks/tools/, velg installasjon tilsvarende ditt operativsystem.

## Konfigurering av pipeline
Åpne filen *pipeline/configs.py*
Sett PIPELINE_NAME til det du ønsker å kalle pipelinen. Unngå spesialtegn, utenom "-" og "_". Husk dette navnet.


Konfigurer om ønskelig øvrige parametere:
- TRAIN_NUM_STEPS (Antall treningssteg)
- TRAINER_EVAL_NUM_STEPS (Antall evalueringssteg)
- EVAL_ACCURACY_THRESHOLD (Minstekrav for godkjent prediksjonssikkerhet 0-1)

## Sette opp infrastruktur i GCP
### Alternativ 1 - Skript
Dette skriptet oppretter en GCS Bucket med nødvendig mappestruktur, laster opp nødvendige filer til bucketen, lager et VPC-nettverk og deployer et Kubernetes-cluster. 

    python pipeline/gcp_infrastructure.py
    
Dersom skriptet kjører fint (uten feilmeldinger utvoer "xxxx eksisterer allerede"), fortsett med [Rulle ut Kubeflow Pipelines](#rulle-ut-kubeflow-pipelines).

### Alternativ 2 - GUI
For bedre kontroll burde det grafiske brukergrensesnittet nyttes (https://console.cloud.google.com/). Velg prosjektet du skal fortsette å arbeide i.

#### Opprett bucket

Navigér til Cloud Storage og opprett bucket:

<img src="media/gcs/create_bucket.png" width="500">

Gi den et navn på formen *"ditt-prosjektnavn*"-navn-på-pipeline

<img src="media/gcs/name-bucket.png" width="500">

Sett lagring til *Multi-region* og *eu*

<img src="media/gcs/bucket-location.png" width="500">

Sett storage class til *Standard*

<img src="media/gcs/storage_class.png" width="500">

Sett Access control til *Fine-grained*

<img src="media/gcs/access-control.png" width="500">

Sett Encryption til *Google-managed encryption key*

<img src="media/gcs/advanced-settings.png" width="500">

Gå inn i bucketen og opprett to mapper, *input* og *output*

<img src="media/gcs/create_folder.png" width="500">

Gå inn i *input/* og last opp mappe (*Upload folder*). Her må mappene */template/data* og */template/modules* lastes opp, med filene som følger. Mappestrukturen skal altså være:

```
|---input/
|      |---data/
|      |     |---data.csv
|      |---modules/
|            |---preprocessor.py
|            |---trainer.py
|---output/
```

#### Opprett nettverk
Google Cloud Console: Gå til *VPC Network* og opprett et nytt nettverk

<img src="media/network/network1.png" width="500">

**Viktig**: Gi nettverket navnet *default* og sett *Subnet creation mode* til *Automatic*

<img src="media/network/network2.png" width="500">

#### Opprett cluster
Google Cloud Console: Gå til *Kubernetes Engine > Clusters* og opprett cluster

<img src="media/cluster/create_cluster.png" width="500">

Velg **Standard** (autopilot er til dags dato ikke støttet). 

Navngi clusteret på formen *<mitt-pipelinenavn>-cluster* og velg *europe-west3-a*
    
<img src="media/cluster/cluster_basics.png" width="500">

Gå til *networking* og velg *default* nettverk og subnettverk

<img src="media/cluster/networking.png" width="500">

#### Kjøre ut KFP på cluster

## Rulle ut Kubeflow Pipelines

Etter nødvendig infrastruktur er opprettet må Kubeflow Pipelines rulles ut på Kubernetes-clusteret. Dette må per nå gjøres manuelt i Google Cloud Console.
Navigér til *AI Platform > Pipelines* og opprett en ny instans

<img src="media/kfp-to-cluster/location.png" width="500">

Clusteret som ble laget tidligere vil automatisk velges som cluster, dersom det er provisjonert ferdig. Vent eventuelt til clusteret er klart før det velges. Velg *default* namespace og gi pipelinen et navn. Deploy uten ekstra konfigurering.

<img src="media/kfp-to-cluster/config.png" width="500">

## Opprett og kjør pipelinen
Sørg for at du står i samme mappe som *kubeflow_runner.py* ligger.

Autentiser kubectl mot clusteret. NB: Bytt **CLUSTER-NAVN** (typisk <mitt-pipelinenavn>-cluster) og **PROSJEKT**

    gcloud container clusters get-credentials CLUSTER-NAVN --zone europe-west3-a --project PROSJEKT

Skru av bruk av caching i pipelinen. Dette unngår en del trøbbel.

    kubectl patch mutatingwebhookconfiguration cache-webhook-default -n default --type='json' -p='[{"op":"replace", "path": "/webhooks/0/rules/0/operations/0", "value": "DELETE"}]'

Finn og notér Kubeflow Pipelines endpoint på formen https://123abc123abc123abc-dot-europe-west1.pipelines.googleusercontent.com.

Dette ligger under AI Platforms > Pipelines. Trykk "Settings" på pipelinen du skal bruke.

<img src="media/kfp-to-cluster/pipeline_settings.png" width="500">

Husk å kopier med hele verdien til "host"

<img src="media/kfp-to-cluster/endpoint.png" width="500">

Opprett pipelinen med TFX. NB: Bytt **ENDPOINT** med endepunktet du skal bruke

    tfx pipeline create --engine=kubeflow --endpoint=ENDPOINT --pipeline-path=kubeflow_runner.py
    
Ved endringer i pipelinen kan følgende kommando kjøres:

    tfx pipeline update --engine=kubeflow --endpoint=ENDPOINT --pipeline-path=kubeflow_runner.py

Start en kjøring av pipelinen. NB: Bytt **PIPELINE_NAME** og **ENGINE**. Dette vil gi en output med en lenke til kjøringen.

    tfx run create --pipeline_name=PIPELINE_NAME --endpoint=ENDPOINT --engine=kubeflow

